"""
The interfaces module defines common interfaces for the major classes 
in the pydsol framework. The use of interfaces (aka abstract base classes
in Python) avoids circular references in the import of modules. 

As an example, the Simulator class refers to a Replication and a Model in 
its initialize method. Both the Replication classes and the Model class
have references to, and use methods from the Simulator class. The interfaces
module that defines the core 'contract' for the Simulator, Model, Experiment,
Replication and Statistics helps to avoid circular references, but also 
defines the core functionalities of these central classes in the pydscol
framework.

Instead of combining all classes in one huge pydsol module with 
thousands of lines of code, the interfaces nicely decouple the definition
of the classes and their implementation, and they avoid circular referencing 
of modules to each other. Think of the use of this particular interface 
module as the .h files in C++.
"""

from abc import ABC, abstractmethod
from typing import TypeVar, Generic, Dict

from pydsol.core.pubsub import EventType
from pydsol.core.utils import get_module_logger

__all__ = [
    "SimulatorInterface",
    "ReplicationInterface",
    "ExperimentInterface",
    "InputParameterInterface",
    "ModelInterface",
    "StatisticsInterface",
    "SimStatisticsInterface",
    "StatEvents",
]

logger = get_module_logger('interfaces')

# ----------------------------------------------------------------------------
# SIMULATOR INTERFACES
# ----------------------------------------------------------------------------

# The TypeVar for time is used for type hinting for simulator time types
TIME = TypeVar("TIME", float, int)


class SimulatorInterface(ABC, Generic[TIME]):
    """
    The SimulatorInterface defines the key methods for any Simulator
    to be used in the pydsol-framework. Different types of Simulators can
    be used, e.g., fixed time increment simulators (time ticks) for ABM
    and for solving differential equations, and variable time increments
    for discrete-event models. Simulators can run as-fast-as-possible or
    be synchronized with the wall clock time, etc. 
    
    Event Types
    -----------
    STARTING_EVENT: EventType
        Will be fired when the simulator has been instructed to start. The 
        actual start might not have happened yet.
    START_EVENT: EventType
        Will be fired when the simulator has actually started.
    STOPPING_EVENT: EventType
        Will be fired when the simulator has been instructed to stop or
        pause. The actual stop might not have happened yet.
    START_EVENT: EventType
        Will be fired when the simulator has actually paused or stopped.
    TIME_CHANGED_EVENT: EventType
        Will be fired when the time of the simulation has changed. This 
        event can be very useful, for instance, to draw time-dependent 
        graphs. 
    """

    STARTING_EVENT: EventType = EventType("STARTING_EVENT")
    START_EVENT: EventType = EventType("START_EVENT")
    STOPPING_EVENT: EventType = EventType("STOPPING_EVENT")
    STOP_EVENT: EventType = EventType("STOP_EVENT")
    TIME_CHANGED_EVENT: EventType = EventType("TIME_CHANGED_EVENT")

    @property
    @abstractmethod
    def name(self) -> str:
        """Return the name of the simulator."""

    @property
    @abstractmethod
    def time_type(self) -> type:
        """Return the time type of the simulator."""

    @property
    @abstractmethod
    def simulator_time(self) -> TIME:
        """Return the current absolute time of the simulator."""

    @property
    @abstractmethod
    def replication(self) -> 'ReplicationInterface':
        """Return the replication with which the simulator has been 
        initialized, or None when initialize has not yet been called."""

    @property
    @abstractmethod
    def model(self) -> 'ModelInterface':
        """Return the model that is being simulated, or None when 
        initialize for a model has not yet been called."""

    @abstractmethod
    def initialize(self, model: 'ModelInterface',
                   replication: 'ReplicationInterface'):
        """Initialize the simulator with a replication for a model."""

    @abstractmethod
    def add_initial_method(self, target, method: str, **kwargs):
        """Add a method call that has to be performed at the end if 
        initialize, and before the model starts. This can, for instance,
        be used to schedule the execution of simulation events before 
        initialize has been called, and solved the problem that,
        for discrete event simulators, the scheduleEvent(...) methods 
        cannot be called before initialize()."""

    @abstractmethod
    def cleanup(self):
        """Clean up after a replication has finished, and prepare for the
        next replication to run."""

    @abstractmethod
    def start(self):
        """Starts the simulator, and fire a START_EVENT when the simulator 
        is started. The start method uses the RunUntil property with a 
        value of the end time of the replication when starting the simulator.
        
        Note
        ----
        Note that when the simulator was already started, an 
        exception will be raised, and no event will be fired."""

    @abstractmethod
    def step(self):
        """Steps the simulator, and fire a START_EVENT before the execution
        of the event, and a STOP_EVENT after the execution of the event to 
        indicate the simulator made a step. 
        
        Note
        ----
        Note that when the simulator is already  running, an exception 
        will be raised, and no event will be fired."""

    @abstractmethod
    def stop(self):
        """Stops or pauses the simulator, and fire a STOP_EVENT when the 
        simulator is stopped. 
        
        Note
        ----
        Note that when the simulator was already stopped, an exception 
        will be raised, and no event will be fired."""

    @abstractmethod
    def run_up_to(self, stop_time: TIME):
        """Runs the simulator up to a certain time; any events at that time, 
        or the solving of the differential equation at that timestep, 
        will not yet be executed."""

    @abstractmethod
    def run_up_to_including(self, stop_time: TIME):
        """Runs the simulator up to a certain time; all events at that time, 
        or the solving of the differential equation at that timestep, 
        will be executed."""

    @abstractmethod
    def is_initialized(self) -> bool:
        """Return whether the simulator has been initialized with a 
        replication for a model."""

    @abstractmethod
    def is_starting_or_running(self) -> bool:
        """Return whether the simulator is starting or has started."""

    @abstractmethod
    def is_stopping_or_stopped(self) -> bool:
        """Return whether the simulator is stopping or has been stopped. 
        This method also returns True when the simulator has not yet been
        initialized, or when the model has not yet started."""


class ReplicationInterface(ABC, Generic[TIME]):
    """
    The ReplicationInterface defines the method that an Replication needs
    to implement. It also defines the events that will be fired to 
    indicate that the execution of a replication on the simulator has
    started, and that a replication on the simulator has ended. A 
    replication provides a start time, warmup time, and duration to the
    simulator and model, and it is related to a unique set of seed values
    for the random streams used in the stochastic simulation. 
    
    Event Types
    -----------
    START_REPLICATION_EVENT: EventType
        Will be fired when the execution of the replication has started.
    END_REPLICATION_EVENT: EventType
        Will be fired when the execution of the replication has completed.
    WARMUP_EVENT: EventType
        Will be fired when the warmup period has been reached, and the
        defined statistics in the model will be cleared.
    """
    START_REPLICATION_EVENT: EventType = EventType("START_REPLICATION_EVENT")
    END_REPLICATION_EVENT: EventType = EventType("END_REPLICATION_EVENT")
    WARMUP_EVENT: EventType = EventType("WARMUP_EVENT")

    @property
    @abstractmethod
    def start_sim_time(self) -> TIME:
        """Return the absolute start time of the replication"""

    @property
    @abstractmethod
    def warmup_sim_time(self) -> TIME:
        """Return the absolute warmup time of the replication"""

    @property
    @abstractmethod
    def end_sim_time(self) -> TIME:
        """Return the absolute end time of the replication"""


class ExperimentInterface(ABC, Generic[TIME]):
    """
    The ExperimentInterface defines the method that an Experiment needs
    to implement. It also defines the events that will be fired to 
    indicate that the execution of an experiment on the simulator has
    started, and that an experiment on the simulator has ended. An 
    experiment consists of a number of replications for the model that 
    will be executed with the same start time, warmup time, and duration. 
    
    Event Types
    -----------
    START_EXPERIMENT_EVENT: EventType
        Will be fired when the execution of the experiment has started.
    END_EXPERIMENT_EVENT: EventType
        Will be fired when the execution of the experiment has completed.
    """
    START_EXPERIMENT_EVENT: EventType = EventType("START_EXPERIMENT_EVENT")
    END_EXPERIMENT_EVENT: EventType = EventType("END_EXPERIMENT_EVENT")


class InputParameterInterface(ABC):
    """
    Input parameters describe different types of input parameters for the 
    model. All parameters for a model are contained in a hierarchical map 
    where successive keys can be retrieved using a dot-notation between 
    the key elements. Suppose a model has two servers: server1 and server2. 
    Each of the servers has an average service time and a number of resources. 
    This can be coded using keys ‘server1.avg_serice_time’, 
    ‘server1.nr_resources’, ‘server2.avg_serice_time’, ‘server2.nr_resources’. 
    This means that the key ‘server1’ contains an `InputParameterMap` with an 
    instance of InputParameterFloat for the service time, and 
    `InputParameterInt` for the number of resources. Readers for the input 
    parameter map can read the model parameters, e.g., from the screen, 
    a web page, an Excel file, a properties file, or a JSON file.
    
    This generic interface describes the minimum set of methods and
    properties that an InputParameter should have. The definition of the
    interface avoids circular references. 
    """

    @property
    @abstractmethod
    def key(self) -> str:
        """
        Return the key of the parameter that can be a part of the 
        dot-notation to uniquely identify the model parameter. The key 
        does not contain the name of the parent. The key is set at time 
        of construction and it is immutable.
        """

    @abstractmethod
    def extended_key(self):
        """
        Return the extended key of this InputParameter including parents 
        with a dot-notation. The name of this parameter is the last entry 
        in the dot notation.
        """

    @property
    @abstractmethod
    def name(self) -> str:
        """
        Returns the concise description of the input parameter, which can 
        be used in a GUI to identify the parameter to the user.
        """

    @property
    @abstractmethod
    def description(self) -> str:
        """
        Returns description or explanation of the InputParameter. For instance,
        an indication of the bounds or the type. This value is purely there 
        for the user interface.
        """

    @property
    @abstractmethod
    def default_value(self) -> object:
        """
        Returns the default (initial) value of the parameter. The actual 
        return type will be defined in subclasses of `InputParameter`. 
        The default value is immutable.
        """

    @property
    @abstractmethod
    def value(self) -> object:
        """
        Returns the actual value of the parameter. The value is initialized
        with default_value and is updated based on user input or data input.
        The actual type will be defined in subclasses of `InputParameter`.
        """

    @abstractmethod
    def set_value(self, value: object):
        """
        Provides a new value for the parameter. The actual type of `value` 
        will be defined in subclasses of `InputParameter`. This is actually
        a method and not a setter property because it can raise errors 
        based on the validity of the value.
        """

    @property
    @abstractmethod
    def display_priority(self) -> float:
        """
        Return the number indicating the order of display of the parameter 
        in the parent parameter map. Floats make it easy to insert an extra 
        parameter between parameters that have already been allocated 
        subsequent integer values.
        """

    @property
    @abstractmethod
    def read_only(self) -> bool:
        """
        Return whether a user is prohibited from changing the value of the
        parameter or not.
        """

    @property
    @abstractmethod
    def parent(self):
        """
        Return the parent map in which the parameter can be retrieved using 
        its  key. Typically, only the root InputParameterMap has no parent, 
        and all other parameters have an InputParameterMap as parent.
        """


class ModelInterface(ABC):
    """
    The ModelInterface defines the minimum set of methods that a simulation
    model in the pydsol-framework should implement. Every model consists of
    the business logic (state transitions initialized in the 
    `construct_model` method), input parameters, output statistics, and a
    reference to the simulator that executes the model. 
    
    The most important method for the Model is the `construct_model` method.
    This method is called for each replication to initialize the model to
    its initial state. The state of the model should be the same every time
    after the `construct_model` method has been called. Constant parts of 
    the model that might be expensive to calculate (e.g., maps, large graphs, 
    information from databases) does not have to be calculated every time 
    in the `construct_model` method, but can be defined once in the `__init__`
    method instead.
    """

    @abstractmethod
    def construct_model(self):
        """
        Code to construct the model logic for each replication. This 
        method is called for each replication to initialize the model to
        its initial state. The state of the model should be the same every 
        time after the `construct_model` method has been called. Constant 
        parts of the model that might be expensive to calculate (e.g., maps, 
        large graphs, information from databases) does not have to be 
        calculated every time in the `construct_model` method, but can be 
        defined once in the `__init__` method instead.
        """

    @property
    @abstractmethod
    def simulator(self) -> SimulatorInterface:
        """Return the simulator for this model."""

    @abstractmethod
    def add_parameter(self, input_parameter):
        """Add an input parameter to the input parameter map."""

    @abstractmethod
    def set_parameter(self, key: str, value: object):
        """set the parameter value of an input parameter."""

    @abstractmethod
    def get_parameter(self, key: str) -> object:
        """return the value of an input parameter."""

    @abstractmethod
    def output_statistics(self) -> Dict[str, "StatisticsInterface"]:
        """return the output statistics map."""

    @abstractmethod
    def add_output_statistic(self, key: str, statistic: "StatisticsInterface"):
        """add an output statistic to the output statistics map."""

    @abstractmethod
    def get_output_statistic(self, key: str) -> "StatisticsInterface":
        """retrieve an output statistic from the output statistics map."""


# ----------------------------------------------------------------------------
# STATISTICS INTERFACES
# ----------------------------------------------------------------------------


class StatisticsInterface(ABC):
    """
    The StatisticsInterface is a tagging interface for statistics classes. 
    It defines the minimum set of method that any statistic in the
    pydsol-framework needs to implement.
    """

    @abstractmethod
    def initialize(self) -> None:
        """Initialize the statistic. This can happen at a the start and/or
        at a simulation replication warmup event."""

    @property
    @abstractmethod
    def name(self) -> str:
        """Return the descriptive name of the statistic."""

    @abstractmethod
    def n(self) -> int:
        """Return the number of observations."""


class SimStatisticsInterface(StatisticsInterface):
    """
    The SimStatisticsInterface is a tagging interface for statistics classes
    that are aware of the Simulator, and that can listen to events such as 
    the WARMUP_EVENT to (re)initialize the statistics. 
    """

    @abstractmethod
    def notify(self, event) -> None:
        """EventListener behavior, so the statistic can be subscribed to 
        events like WARMUP_EVENT and END_REPLICATION_EVENT.  
        """

    @property
    @abstractmethod
    def simulator(self) -> SimulatorInterface:
        """Return the simulator."""


# ----------------------------------------------------------------------------
# STATISTICS EVENTS
# ----------------------------------------------------------------------------


class StatEvents:
    """
    StatEvents contains the set of events that different statistics in
    the statistics module can fire. To avoid circular references between the
    statistics module that fires the events, and modules that listen to 
    events, all events are defines in one place as `StatEvents.XXX_EVENT`.
    
    The events that can be used are listed below.
    """

    DATA_EVENT: EventType = EventType("DATA_EVENT")
    """
    The DATA_EVENT is the incoming event for EventBased statistics that 
    contains a new value for the statistics. The payload is a single float. 
    This event can be used by the EventBasedCounter and EventBasedTally 
    and its subclasses. The event is fired from outside to these statistics.
    """

    WEIGHT_DATA_EVENT: EventType = EventType("WEIGHT_DATA_EVENT")
    """The WEIGHT_DATA_EVENT is the incoming event for weighted EventBased 
    statistics that contains a new weight-value pair for the statistics. 
    The payload is a tuple (weight, value). This event can be used by the 
    EventBasedWeightedTally and its subclasses.The event is fired from 
    outside to the statistics."""

    TIMESTAMP_DATA_EVENT: EventType = EventType("TIMESTAMP_DATA_EVENT")
    """The TIMESTAMP_DATA_EVENT is the incoming event for weighted EventBased 
    statistics that contains a new timestamp-value pair for the statistics. 
    The payload is a tuple (timestamp, value). This event can be used by the 
    EventBasedTimestampWeightedTally and its subclasses.The event is fired 
    from outside to the statistics."""

    INITIALIZED_EVENT: EventType = EventType("INITIALIZED_EVENT")
    """INITIALIZED_EVENT indicates that the statistic has been
    (re)initialized, and all counters have been reset to the original values.
    The event is fired by the statistic to its listeners."""

    OBSERVATION_ADDED_EVENT: EventType = EventType("OBSERVATION_ADDED_EVENT")
    """OBSERVATION_ADDED_EVENT indicates that an observation has been 
    received, and contains the value of the observation as the payload.
    For weight-bsaed and timestamp-based observations, the payload is a tuple.
    For value-based statistics, the payload is just the (float) value. 
    The event is fired by the statistic to its listeners."""

    N_EVENT: EventType = EventType("N_EVENT")
    """N_EVENT indicates that the number of observations has been increased,
    and contains the new number of observation as the payload. 
    The event is fired by the statistic to its listeners."""

    COUNT_EVENT: EventType = EventType("COUNT_EVENT")
    """COUNT_EVENT is an event of the Counter statistic and is subclasses.
    The event indicates that the count has changed, and contains the new 
    count as the payload. The event is fired by the statistic to its 
    listeners."""

    MIN_EVENT: EventType = EventType("MIN_EVENT")
    """MIN_EVENT indicates that the minimum of the observations has changed,
    and contains the new minimum as the payload. This event is fired by the 
    EventBasedTally, its Weighted and Timestamped variant, and its subclasses 
    to the listeners."""

    MAX_EVENT: EventType = EventType("MAX_EVENT")
    """MAX_EVENT indicates that the maximum of the observations has changed,
    and contains the new maximum as the payload. This event is fired by the 
    EventBasedTally, its Weighted and Timestamped variants, and their 
    subclasses to the listeners."""

    SUM_EVENT: EventType = EventType("SUM_EVENT")
    """SUM_EVENT indicates that the sum of the observations has changed,
    and contains the new sum as the payload. This event is fired by the 
    EventBasedTally and its subclasses to the listeners."""

    MEAN_EVENT: EventType = EventType("MEAN_EVENT")
    """MEAN_EVENT indicates that the mean of the observations has changed, 
    and contains the new mean as the payload. This event is fired by the 
    EventBasedTally and its subclasses to the listeners."""

    POPULATION_STDEV_EVENT: EventType = EventType("POPULATION_STDEV_EVENT")
    """POPULATION_STDEV_EVENT indicates that the population standard
    deviation of the observations has changed, and contains the new standard
    deviation as the payload. This event is fired by the EventBasedTally and 
    its subclasses to the listeners."""

    POPULATION_VARIANCE_EVENT: EventType = EventType("POPULATION_VARIANCE_EVENT")
    """POPULATION_VARIANCE_EVENT indicates that the population variance
    of the observations has changed, and contains the new variance as the 
    payload. This event is fired by the EventBasedTally and its subclasses 
    to the listeners."""

    POPULATION_SKEWNESS_EVENT: EventType = EventType("POPULATION_SKEWNESS_EVENT")
    """POPULATION_SKEWNESS_EVENT indicates that the population skewness
    of the observations has changed, and contains the new skewness as the 
    payload. This event is fired by the EventBasedTally and its subclasses 
    to the listeners."""

    POPULATION_KURTOSIS_EVENT: EventType = EventType("POPULATION_KURTOSIS_EVENT")
    """POPULATION_KURTOSIS_EVENT indicates that the population kurtosis
    of the observations has changed, and contains the new kurtosis as the 
    payload. This event is fired by the EventBasedTally and its subclasses 
    to the listeners."""

    POPULATION_EXCESS_K_EVENT: EventType = EventType("POPULATION_EXCESS_K_EVENT")
    """POPULATION_EXCESS_K_EVENT indicates that the population exces kurtosis
    of the observations has changed, and contains the new excess kurtosis as 
    the payload. This event is fired by the EventBasedTally and its subclasses 
    to the listeners."""

    SAMPLE_STDEV_EVENT: EventType = EventType("SAMPLE_STDEV_EVENT")
    """SAMPLE_STDEV_EVENT indicates that the sample standard deviation of 
    the observations has changed, and contains the new standard deviation 
    as the payload. This event is fired by the EventBasedTally and its 
    subclasses to the listeners."""

    SAMPLE_VARIANCE_EVENT: EventType = EventType("SAMPLE_VARIANCE_EVENT")
    """SAMPLE_VARIANCE_EVENT indicates that the sample variance of the 
    observations has changed, and contains the new variance as the payload. 
    This event is fired by the EventBasedTally and its subclasses to the 
    listeners."""

    SAMPLE_SKEWNESS_EVENT: EventType = EventType("SAMPLE_SKEWNESS_EVENT")
    """SAMPLE_SKEWNESS_EVENT indicates that the sample skewness of the 
    observations has changed, and contains the new skewness as the payload. 
    This event is fired by the EventBasedTally and its subclasses to the 
    listeners."""

    SAMPLE_KURTOSIS_EVENT: EventType = EventType("SAMPLE_KURTOSIS_EVENT")
    """SAMPLE_KURTOSIS_EVENT indicates that the sample kurtosis of the 
    observations has changed, and contains the new kurtosis as the payload. 
    This event is fired by the EventBasedTally and its subclasses to the 
    listeners."""

    SAMPLE_EXCESS_K_EVENT: EventType = EventType("SAMPLE_EXCESS_K_EVENT")
    """SAMPLE_EXCESS_K_EVENT indicates that the sample excess kurtosis of the 
    observations has changed, and contains the new excess kurtosis as the 
    payload. This event is fired by the EventBasedTally and its subclasses 
    to the listeners."""

    WEIGHTED_SUM_EVENT: EventType = EventType("WEIGHTED_SUM_EVENT")
    """WEIGHTED_SUM_EVENT indicates that the weighted sum of the observations 
    has changed, and contains the new weighted sum as the payload. This event 
    is fired by the EventBasedWeightedTally and its subclasses to the 
    listeners."""

    WEIGHTED_MEAN_EVENT: EventType = EventType("WEIGHTED_MEAN_EVENT")
    """WEIGHTED_MEAN_EVENT indicates that the weighted mean of the 
    observations has changed, and contains the new weighted mean 
    as the payload. This event is fired by the EventBasedWeightedTally and 
    its subclasses to the listeners."""

    WEIGHTED_POPULATION_STDEV_EVENT: EventType = \
        EventType("WEIGHTED_POPULATION_STDEV_EVENT")
    """WEIGHTED_POPULATION_STDEV_EVENT indicates that the weighted population 
    standard deviation of the observations has changed, and contains the 
    new weighted standard deviation as the payload. This event is fired by 
    the EventBasedWeightedTally and its subclasses to the listeners."""

    WEIGHTED_POPULATION_VARIANCE_EVENT: EventType = \
        EventType("WEIGHTED_POPULATION_VARIANCE_EVENT")
    """WEIGHTED_POPULATION_VARIANCE_EVENT indicates that the weighted 
    population variance of the observations has changed, and contains the 
    new weighted variance as the payload. This event is fired by the 
    EventBasedWeightedTally and its subclasses to the listeners."""

    WEIGHTED_SAMPLE_STDEV_EVENT: EventType = \
        EventType("WEIGHTED_SAMPLE_STDEV_EVENT")
    """WEIGHTED_SAMPLE_STDEV_EVENT indicates that the weighted sample 
    standard deviation of the observations has changed, and contains the 
    new weighted standard deviation as the payload. This event is fired by 
    the EventBasedWeightedTally and its subclasses to the listeners."""

    WEIGHTED_SAMPLE_VARIANCE_EVENT: EventType = \
        EventType("WEIGHTED_SAMPLE_VARIANCE_EVENT")
    """WEIGHTED_SAMPLE_VARIANCE_EVENT indicates that the weighted sample 
    variance of the observations has changed, and contains the new weighted 
    variance as the payload. This event is fired by the EventBasedWeightedTally 
    and its subclasses to the listeners."""



"""
This module contains generic utilities for the pydsol code. 
"""
import logging
import math
import sys

__all__ = [
    "DSOLError",
    "get_module_logger",
    "sign",
    "erf_inv",
    "beta",
]


class DSOLError(Exception):
    """General Exception class for pydsol."""
    pass


logging.basicConfig(level=logging.CRITICAL,
                    format='%(levelname)s: %(module)s.%(funcName)s: %(message)s')


def get_module_logger(mod_name: str, level=logging.CRITICAL):
    logger = logging.getLogger(mod_name)
    if not logger.hasHandlers():
        handler = logging.StreamHandler(stream=sys.stdout)
        logger.addHandler(handler)
    msg_format = '%(levelname)s: %(module)s.%(funcName)s: %(message)s'
    formatter = logging.Formatter(fmt=msg_format)
    for handler in logger.handlers:
        handler.setFormatter(formatter)
    logger.setLevel(level)
    return logger


logger = get_module_logger('utils')


def sign(x: float) -> float:
    """
    Return the sign of x. Analogous to other programming languages, the
    following convention is used:
    
    * return -1  for negative x
    * return +1  for positive x
    * return 0   when x is zero
    * return nan when x is nan
    """
    if math.isnan(x):
        return math.nan
    if x > 0:
        return 1.0
    if x < 0:
        return -1.0
    return 0.0


def erf_inv(y: float) -> float:
    """
    Approximates the inverse error function (erf) based on the C-algorithm
    at http://www.naic.edu/~jeffh/inverse_cerf.c.
    
    Raises
    ------
    TypeError
        when y is not a number
    ValueError
        when y is not between -1 and 1 (inclusive)
    """
    if not isinstance(y, (float, int)):
        raise TypeError(f"Parameter y {y} is not a number")
    if not -1.0 <= y <= 1.0:
        raise ValueError(f"Parameter y {y} not between 0 and 1 (inclusive)")

    ax = abs(y)

    if ax <= 0.75:
        # This approximation, taken from Table 10 of Blair et al., is valid
        # for |x|<=0.75 and has a maximum relative error of 4.47 x 10^-8.
        p = (-13.0959967422, 26.785225760, -9.289057635)
        q = (-12.0749426297, 30.960614529, -17.149977991, 1.00000000)
        t = ax * ax - 0.75 * 0.75
        r = (ax * (p[0] + t * (p[1] + t * p[2]))
             / (q[0] + t * (q[1] + t * (q[2] + t * q[3]))))

    elif 0.75 <= ax <= 0.9375:
        # This approximation, taken from Table 29 of Blair et al., is valid 
        # for .75<=|x|<=.9375 and has a maximum relative error of 4.17 x 10^-8.
        p = (-.12402565221, 1.0688059574, -1.9594556078, .4230581357)
        q = (-.08827697997, .8900743359, -2.1757031196, 1.0000000000)
        t = ax * ax - 0.9375 * 0.9375
        r = (ax * (p[0] + t * (p[1] + t * (p[2] + t * p[3])))
             / (q[0] + t * (q[1] + t * (q[2] + t * q[3]))))

    elif 0.9375 <= ax <= (1.0 - 1.0e-9):
        # This approximation, taken from Table 50 of Blair et al., is valid 
        # for .9375<=|x|<=1-10^-100 and has a maximum relative error of 
        # 2.45 x 10^-8.
        p = (.1550470003116, 1.382719649631, .690969348887, -1.128081391617,
             .680544246825, -.16444156791)
        q = (.155024849822, 1.385228141995, 1.000000000000)
        t = 1.0 / math.sqrt(-math.log(1.0 - ax))
        r = ((p[0] / t + p[1] + t * (p[2] + t * (p[3] + t * (p[4] + t * p[5]))))
             / (q[0] + t * (q[1] + t * (q[2]))))

    else:
        r = math.inf

    return sign(y) * r


def beta(z: float, w: float) -> float:
    """
    Calculates Beta(z, w) where Beta(z, w) = Gamma(z) * Gamma(w) / Gamma(z + w).
    
    Raises
    ------
    TypeError
        when z or w are not numbers 
    ValueError
        when z < 0 or w < 0
    """
    if not isinstance(w, (float, int)):
        raise TypeError(f"Parameter w {w} is not a number")
    if not isinstance(z, (float, int)):
        raise TypeError(f"Parameter z {z} is not a number")
    if z < 0 or w < 0:
        raise ValueError(f"Parameter z {z} or w {w} < 0")
    return math.exp(math.lgamma(z) + math.lgamma(w) - math.lgamma(z + w))



"""
the pubsub module contains a number of base classes for the publish/subscribe 
pattern as used in pydsol. The package is based on the Java 
implementation of DSOL (Distributed Simulation Object Library), 
first documented in 2002. Information about the Java and Python DSOL 
libraries can be found at https://simulation.tudelft.nl. Specific 
information about the publish/subscribe implementation in Java can be 
found in the djutils project (Delft Java Utilities) at https://djutils.org

Typically, implemented listener classes inherit from the abstract base class
EventListener (possibly through mixin), and classes that produce events 
inherit from the abstract base class EventProducer (possibly using multiple 
inheritance). 

The whole idea of publish/subscribe is that you decouple the producer of
information and the possible receiver(s), which otherwise has to be
hardcoded, making information exchange between objects very inflexible. 
With pub/sub, the producer will notify only subscribed objects of the 
content in the event. The receivers can change dynamically over time, by 
subscribing or unsubscribing for the receipt of certain EventTypes. We call 
the process of producing an event by the publisher "firing" of an event 
with the fire() method; receipt by the receiver is done in the notify() 
method. 

Note that Events are completely different from SimEvents. Events are used
in publish/subscribe; SimEvents contain information about code to be 
executed at a later point in time.
"""

from abc import ABC, abstractmethod
import inspect
from typing import Type, Optional, Any, Union, Set, Dict
from pydsol.core.utils import get_module_logger

__all__ = [
    "EventError",
    "EventType",
    "Event",
    "TimedEvent",
    "EventListener",
    "EventProducer",
    ]

logger = get_module_logger('pubsub')


class EventError(Exception):
    """General Exception for working with Events and publish/subscribe"""
    pass


class EventType:
    """
    EventType is a strongly typed identifier for an event, which can contain
    additional information about the event itself such as metadata. The
    EventType is typically instantiated as a static variable in a class.
    
    The implementation of the `EventType` takes care that there will not be
    a problem with duplicate names for EventTypes. As part of the `EventType`,
    the defining class is coded as well to make the `EventType` unique. So,
    in the above example, the name "Producer.PRODUCTION" is stored internally
    as the unique name for the `EventType`.
    
    Example
    -------
        .. code-block:: python
        
           class Producer
               PRODUCTION_EVENT = EventType("PRODUCTION")
        
        The event type can then be used anywhere in the code as 
        `Producer.PRODUCTION_EVENT`. Note that an `Event` and a `SimEvent` 
        are two completely different concepts: Events are for the 
        publish/subscribe pattern, SimEvents are to store deferred method 
        information.
    """
    
    # set of the defined types to check for name clashes
    # each item in the set has the form class_name.event_type_name
    __defined_types: Set[str] = set()
    def __init__(self, name: str, metadata: Dict[str, Type]=None):
        """
        Instantiate a new EventType, usually in a static manner. 
        
        Example
        -------
            .. code-block:: python
        
               class Producer
                   PRODUCTION_EVENT = EventType("PRODUCTION")
            
            The event type can then be used anywhere in the code as 
            `Producer.PRODUCTION_EVENT`. 
        
        Parameters
        ----------
        name : str
            the human readable name of the event
        metadata : dict[str, Type], optional
            a dict with the metadata, defining the structure of the payload 
            in the event, as pairs of the name to be used in the dict and
            the expected type of the data field. When metadata is None
            or undefined, the payload of the event can be anything. When
            metadata IS defined, the payload has to be a dict.
            
        Raises
        ------
        EventError
            when name is not a str, or defining_class is not a type
        EventError
            when there was already an event defined with this name 
            in the defining_class
        EventError
            when the metadata does not consist of [str, Type] pairs
            
        """
        if not isinstance(name, str):
            raise EventError("name {name} is not a str")
        self._defining_class: str = inspect.stack()[1][0].f_code.co_name
        self._name = name
        key = self._defining_class + "." + self._name;
        if key in EventType.__defined_types:
            raise EventError(f"EventType {name} already defined")
        EventType.__defined_types.add(key)
        if metadata is not None:
            for key in metadata.keys():
                if not isinstance(key, str):
                    raise EventError("metadata {metadata} key not a str")
                if not isinstance(metadata[key], type):
                    raise EventError("metadata {metadata} value not a type")
        self._metadata = metadata
    
    @property
    def name(self):
        """Return the human readable name of the event type."""
        return self._name
    
    @property
    def defining_class(self):
        """Return the name of the defining class.
        
        The defining class is the class in which the event type has been 
        defined"""
        return self._defining_class
    
    @property
    def metadata(self) -> Optional[Dict[str, Type]]:
        """Return the metadata dict or None.
        
        The metadata dict contains the expected structure of the payload 
        dict of the event in case it is defined. metadata can be None, 
        in which case the payload does not have to be a dict, and can have 
        any structure."""
        return self._metadata
    
    def __str__(self):
        return f"EventType[{self._defining_class}.{self._name}]"
    
    def __repr__(self):
        return f"EventType[{self._defining_class}.{self._name} " \
            +"metadata={self._metadata}]"
    

class Event:
    """
    An event is an object with a payload that is to be sent between 
    a producer and zero or more listeners. In a sense, the Event is the
    "envelope" of the content. 
    """

    def __init__(self, event_type: EventType, content, check:bool=True):
        """
        Instantiate an event with content. Events are strongly typed
        using a (usually static) instance of the EventType class, to
        distinguish different types of events from each other.
        
        Parameters
        ----------
        event_type : EventType
            a reference to a (usually static) event type for identification
        content
            the payload of the event, which can be of any type; common types
            are list, dict, or simple types
        check : bool, optional
            whether to check the fields in the content in case the 
            event_type has metadata; the check whether the content is a
            dict is always checked when there is metadata 
            
        Raises
        ------
        EventError
            if event_type is not an EventType
        EventError
            if the specified metadata and the content is not a dict
        EventError
            if the dict content is not consistent with the EventType metadata
        """
        if not isinstance(event_type, EventType):
            raise EventError("event_type is not an instance of EventType")
        self._event_type = event_type
        self._content = content
        if event_type.metadata is not None:
            if not isinstance(content, dict):
                raise EventError("event_type defined metadata but content "
                    +"is not specified as a dict")
            if check:
                if len(event_type.metadata) != len(content):
                    raise EventError("metadata length but consistent with "
                        +"content length")
                for key in event_type.metadata.keys():
                    if dict(content).get(key, None) == None:
                        raise EventError("metadata requited key '{key}' not "
                        +"found in content dict")
                    if not isinstance(dict(content).get(key),
                                      event_type.metadata.get(key)):
                        raise EventError("metadata requited key '{key}' to "
                        +"be of type {event_type.metadata.get(key).__name__}" 
                        +"but instead got {dict(content).get(key)}")
    
    @property
    def event_type(self) -> EventType:
        """Returns the (usually static) event type for identificatio.n"""
        return self._event_type
        
    @property
    def content(self) -> Any:
        """Returns the payload of the event; can be of any type."""
        return self._content
    
    def __str__(self):
        c = ""
        try:
            c = str(self._content)
        except:
            c = f"[cannot print {type(self._content).__name__}"
        return f"Event[{self._event_type.name}: {c}]"
    
    def __repr__(self):
        c = ""
        try:
            c = str(self._content)
        except:
            c = f"[cannot print {type(self._content).__name__}"
        return f"Event[{self._event_type.name}: {c}]"


class TimedEvent(Event):
    """
    A TimedEvent is an event with a time identifier (int, float).
    
    The time identifier is not specified as an extra element in the payload,
    but rather as an extra attribute of the Event. The constructor explicitly 
    demands for the specification of a timestamp, typically the simulator time. 
    Like an event, it is an object with a payload that is to be sent between 
    a producer and zero or more listeners. In a sense, the TimedEvent is the
    timestamped "envelope" of the content. 
    """

    def __init__(self, timestamp: Union[float, int], event_type: EventType,
                 content, check:bool=True):
        """
        Instantiate a timed event with content. 
        
        TimedEvents are strongly typed using a (usually static) instance of 
        the EventType class, to distinguish different types of events from 
        each other. Furthermore, the timestamp indicates when the event was 
        fired. Typically the value of the timestamp is the simulator time.
        
        Parameters
        ----------
        timestamp : int or float
            the timestamp of the event, typically the simulator time
        event_type : EventType
            a reference to a (usually static) event type for identification
        content
            the payload of the event, which can be of any type; common types
            are list, dict, or simple types
        check : bool, optional
            whether to check the fields in the content in case the 
            event_type has metadata; the check whether the content is a
            dict is always checked when there is metadata 
            
        Raises
        ------
        EventError
            if timestamp is not of type int or float
        EventError
            if event_type is not an EventType
        EventError
            if the EventType specified metadata and the content is not a dict
        EventError
            if the dict content is not consistent with the EventType metadata
        """
        if not isinstance(timestamp, (int, float)):
            raise EventError("timestamp is not an int or a float")
        self._timestamp = timestamp
        super().__init__(event_type, content, check)
    
    @property
    def timestamp(self) -> Union[int, float]:
        """Returns the timestamp of the event; typically the simulator time."""
        return self._timestamp
    
    def __str__(self):
        c = ""
        try:
            c = str(self._content)
        except:
            c = f"[cannot print {type(self._content).__name__}"
        return f"TimedEvent[t={self.timestamp}, {self._event_type.name}: {c}]"
    
    def __repr__(self):
        c = ""
        try:
            c = str(self._content)
        except:
            c = f"[cannot print {type(self._content).__name__}"
        return f"TimedEvent[t={self.timestamp}, {self._event_type.name}: {c}]"


class EventListener(ABC):
    """
    The EventListener abstract class defines the behavior of an event subscriber.
    
    The EventListener is an interface for a class that needs to be able to
    receive events from one or more EventProducers. In order to receive
    events, a listener has to implement the notify() method. In the
    notify() method, events can be tested for their EventType with if-elif
    statements, and then the corresponding content can be acted upon.
    
    Its most important method is notify(event) that is called from the 
    EventProducer (using the fier(event) method) to handle the Event.
    """

    @abstractmethod
    def notify(self, event: Event):
        """Handle an event received from an EventProducer"""


class EventProducer:
    """
    EventProducer is an abstract class defining the producer behavior.
    
    The EventProducer class acts as a superclass for classes that need
    to fire events to an unknown and possibly varying number of subscribers
    (also called listeners). The main methods that can be called on the
    EventProducer are: add_listener and remove_listener. In addition, the
    logic of the class that extends the base EventProducer class calls the
    fire(event_type, content) method to notify the listener(s) (if any).
    
    The most important private attribute of the EventProducer is 
    ``_listeners: dict[EventType, list[EventListener]]``. This structure
    Maps the EventType to a list of listeners for that EventType.
    Note that this is a list to make behavior reproducible: we want
    events to subscribers to be fired in the same order when replicating
    the model run. The dictionary is ordered (unsorted) in Python 3.7+, 
    and the list is reproducible. A ``dict[EventType, set[EventListener]]`` 
    would not be reproducible, since the set is unordered.   
    """

    def __init__(self):
        """Instantiate the EventProducer, and initialize the empty 
        listener data structure"""
        self._listeners: dict[EventType, list[EventListener]] = dict()
        
    def add_listener(self, event_type: EventType, listener: EventListener):
        """
        Add an EventListener to this EventProducer for a given EventType.
        If the listener already is registered for this EventType, this will
        be ignored.
        
        Parameters
        ----------
        event_type : EventType
            the EventType for which this listener subscribes
        listener : EventListener
            the subscriber to register for the provided Eventtype
            
        Raises
        ------
        EventError
            if any of the arguments is of the wrong type
        """
        if not isinstance(event_type, EventType):
            raise EventError("event_type should be an EventType")
        if not isinstance(listener, EventListener):
            raise EventError("listener should be an EventListener")
        if event_type not in self._listeners:
            self._listeners[event_type] = []
        if listener not in self._listeners[event_type]:
            self._listeners[event_type].append(listener)
    
    def remove_listener(self, event_type: EventType, listener: EventListener):
        """
        Remove an EventListener of this EventProducer for a given EventType.
        If the listener is not registered for this EventType, this will
        be ignored.
        
        Parameters
        ----------
        event_type : EventType
            the EventType for which this listener unsubscribes
        listener : EventListener
            the subscriber to remove for the provided Eventtype
            
        Raises
        ------
        EventError
            if any of the arguments is of the wrong type
        """
        if not isinstance(event_type, EventType):
            raise EventError("event_type should be an EventType")
        if not isinstance(listener, EventListener):
            raise EventError("listener should be an EventListener")
        if event_type in self._listeners:
            if listener in self._listeners[event_type]:
                self._listeners[event_type].remove(listener)
                if len(list(self._listeners[event_type])) == 0:
                    del self._listeners[event_type]
    
    def remove_all_listeners(self, event_type:EventType=None,
                        listener:EventListener=None):
        """
        Remove an EventListener (if given) for a provided EventType (if given)
        for this EventProducer. It is no problem if there are no matches. 
        There are four situations:
        
        event_type == None and listener == None
            all listeners for all event types are removed
        event_type == None and listener is specified
            the listener is removed for any event for which it was registered
        event_type is specified and listener == None
            all listeners are removed for the given event_type
        event_type and listener are both specified
            the listener for the given event type is removed, if it was
            registered; in essence this is the same as remove_listener 
        
        Parameters
        ----------
        event_type : EventType, optional
            the EventType for which this listener unsubscribes
        listener : EventListener, optional
            the subscriber to remove for the provided EventType
        
        Raises
        ------
        EventError
            if any of the arguments is of the wrong type
        """
        if not (event_type == None or isinstance(event_type, EventType)):
            raise EventError("event_type should be an EventType or None")
        if not (listener == None or isinstance(listener, EventListener)):
            raise EventError("listener should be an EventListener or None")
        if event_type == None:
            if listener == None:
                self._listeners.clear()
            else:
                # a list() is used to avoid concurrent modification error
                for et in list(self._listeners.keys()):
                    self.remove_listener(et, listener)
        else:
            if listener == None:
                if event_type in self._listeners:
                    del self._listeners[event_type]
            else:
                self.remove_listener(event_type, listener)
    
    def has_listeners(self) -> bool:
        """indicate whether this producer has any listeners or not"""
        return len(self._listeners) > 0
    
    def fire_event(self, event: Event):
        """
        fire this event to the subscribed listeners for the EventType of
        the event.
        
        Parameters
        ----------
        event : Event
            the event to fire to the subscribed listeners
        
        Raises
        ------
        EventError
            if the event is not of the right type
        """
        if not isinstance(event, Event):
            raise EventError("event {event} not of type Event")
        logger.debug("fire %s to %s", event,
                     self._listeners.get(event.event_type))
        if event.event_type not in self._listeners:
            return
        # a copy() is used to avoid concurrent modification error in case 
        # the notification would unsubscribe a listener to this event (!)
        for listener in self._listeners.get(event.event_type).copy():
            listener.notify(event)
 
    def fire(self, event_type: EventType, content, check: bool=True):
        """
        construct an event based on the arguments and fire this event 
        to the subscribed listeners for the event_type
        
        Parameters
        ----------
        event_type : EventType
            a reference to a (usually static) event type for identification
        content
            the payload of the event, which can be of any type; common types
            are list, dict, or simple types
        check : bool, optional
            whether to check the fields in the content in case the 
            event_type has metadata; the check whether the content is a
            dict is always checked when there is metadata 
            
        Raises
        ------
        EventError
            if event_type is not an EventType
        EventError
            if the EventType specified metadata and the content is not a dict
        EventError
            if the dict content is not consistent with the EventType metadata
        """
        event = Event(event_type, content, check)
        self.fire_event(event)
    
    def fire_timed_event(self, timed_event: TimedEvent):
        """
        fire this timed_event to the subscribed listeners for the EventType 
        of the event.
        
        Parameters
        ----------
        event : TimedEvent
            the timed_event to fire to the subscribed listeners
        
        Raises
        ------
        EventError
            if the timed_event is not of the right type
        """
        if not isinstance(timed_event, TimedEvent):
            raise EventError("event {event} not of type TimedEvent")
        logger.debug("fire %s to %s", timed_event,
                     self._listeners.get(timed_event.event_type))
        if timed_event.event_type not in self._listeners:
            return
        # a copy() is used to avoid concurrent modification error in case 
        # the notification would unsubscribe a listener to this event (!)
        for listener in self._listeners.get(timed_event.event_type).copy():
            listener.notify(timed_event)

    def fire_timed(self, time: Union[int, float], event_type: EventType,
                   content, check: bool=True):
        """
        construct a timed event based on the arguments and fire this event 
        to the subscribed listeners for the event_type
        
        Parameters
        ----------
        timestamp : int or float
            the timestamp of the event, typically the simulator time
        event_type : EventType
            a reference to a (usually static) event type for identification
        content
            the payload of the event, which can be of any type; common types
            are list, dict, or simple types
        check : bool, optional
            whether to check the fields in the content in case the 
            event_type has metadata; the check whether the content is a
            dict is always checked when there is metadata 
            
        Raises
        ------
        EventError
            if timestamp is not of type int or float
        EventError
            if event_type is not an EventType
        EventError
            if the EventType specified metadata and the content is not a dict
        EventError
            if the dict content is not consistent with the EventType metadata
        """
        timed_event = TimedEvent(time, event_type, content, check)
        self.fire_timed_event(timed_event)
